@@ -10,6 +10,7 @@
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
+import json
 import math
 import os
 import shutil
@@ -20,6 +21,7 @@ import time
 import uuid
 from datetime import datetime, timedelta
 
+import pytest
 from dateutil.tz import tzlocal
 
 from botocore import UNSIGNED
@@ -44,6 +46,7 @@ from botocore.exceptions import (
 )
 from botocore.session import Session
 from botocore.stub import Stubber
+from botocore.tokens import SSOTokenProvider
 from botocore.utils import datetime2timestamp
 from tests import (
     BaseEnvVar,
@@ -56,6 +59,9 @@ from tests import (
     unittest,
 )
 
+TIME_IN_ONE_HOUR = datetime.utcnow() + timedelta(hours=1)
+TIME_IN_SIX_MONTHS = datetime.utcnow() + timedelta(hours=4320)
+
 
 class TestCredentialRefreshRaces(unittest.TestCase):
     def assert_consistent_credentials_seen(self, creds, func):
@@ -246,9 +252,10 @@ class TestAssumeRole(BaseAssumeRoleTest):
         mock_instance.CANONICAL_NAME = provider_cls.CANONICAL_NAME
         return mock_instance
 
-    def create_session(self, profile=None):
+    def create_session(self, profile=None, sso_token_cache=None):
         session = StubbedSession(profile=profile)
-
+        if not sso_token_cache:
+            sso_token_cache = JSONFileCache(self.tempdir)
         # We have to set bogus credentials here or otherwise we'll trigger
         # an early credential chain resolution.
         sts = session.create_client(
@@ -271,7 +278,7 @@ class TestAssumeRole(BaseAssumeRoleTest):
             ),
             profile_provider_builder=ProfileProviderBuilder(
                 session,
-                sso_token_cache=JSONFileCache(self.tempdir),
+                sso_token_cache=sso_token_cache,
             ),
         )
         stubber = session.stub('sts')
@@ -565,7 +572,7 @@ class TestAssumeRole(BaseAssumeRoleTest):
         with self.assertRaises(InvalidConfigError):
             session.get_credentials()
 
-    def test_sso_source_profile(self):
+    def test_sso_source_profile_legacy(self):
         token_cache_key = 'f395038c92f1828cbb3991d2d6152d326b895606'
         cached_token = {
             'accessToken': 'a.token',
@@ -624,6 +631,68 @@ class TestAssumeRole(BaseAssumeRoleTest):
         }
         self.assertEqual(kwargs, expected_kwargs)
 
+    def test_sso_source_profile(self):
+        token_cache_key = '32096c2e0eff33d844ee6d675407ace18289357d'
+        cached_token = {
+            'accessToken': 'C',
+            'expiresAt': TIME_IN_ONE_HOUR.strftime('%Y-%m-%dT%H:%M:%SZ'),
+        }
+        temp_cache = JSONFileCache(self.tempdir)
+        temp_cache[token_cache_key] = cached_token
+        config = (
+            '[profile A]\n'
+            'role_arn = arn:aws:iam::123456789:role/RoleA\n'
+            'source_profile = B\n'
+            '[profile B]\n'
+            'sso_session = C\n'
+            'sso_role_name = SSORole\n'
+            'sso_account_id = 1234567890\n'
+            '[sso-session C]\n'
+            'sso_region = us-east-1\n'
+            'sso_start_url = https://test.url/start\n'
+        )
+        self.write_config(config)
+
+        session, sts_stubber = self.create_session(
+            profile='A', sso_token_cache=temp_cache
+        )
+        client_config = Config(
+            region_name='us-east-1',
+            signature_version=UNSIGNED,
+        )
+        sso_stubber = session.stub('sso', config=client_config)
+        sso_stubber.activate()
+        # The expiration needs to be in milliseconds
+        expiration = datetime2timestamp(self.some_future_time()) * 1000
+        sso_role_creds = self.create_random_credentials()
+        sso_role_response = {
+            'roleCredentials': {
+                'accessKeyId': sso_role_creds.access_key,
+                'secretAccessKey': sso_role_creds.secret_key,
+                'sessionToken': sso_role_creds.token,
+                'expiration': int(expiration),
+            }
+        }
+        sso_stubber.add_response('get_role_credentials', sso_role_response)
+
+        expected_creds = self.create_random_credentials()
+        assume_role_response = self.create_assume_role_response(expected_creds)
+        sts_stubber.add_response('assume_role', assume_role_response)
+
+        actual_creds = session.get_credentials()
+        self.assert_creds_equal(actual_creds, expected_creds)
+        sts_stubber.assert_no_pending_responses()
+        # Assert that the client was created with the credentials from the
+        # SSO get role credentials response
+        self.assertEqual(self.mock_client_creator.call_count, 1)
+        _, kwargs = self.mock_client_creator.call_args_list[0]
+        expected_kwargs = {
+            'aws_access_key_id': sso_role_creds.access_key,
+            'aws_secret_access_key': sso_role_creds.secret_key,
+            'aws_session_token': sso_role_creds.token,
+        }
+        self.assertEqual(kwargs, expected_kwargs)
+
     def test_web_identity_credential_source_ignores_env_vars(self):
         token_path = os.path.join(self.tempdir, 'token')
         with open(token_path, 'w') as token_file:
@@ -929,3 +998,148 @@ class TestSTSRegional(BaseAssumeRoleTest):
             self.assertEqual(
                 stubber.requests[0].url, 'https://sts.us-west-2.amazonaws.com/'
             )
+
+
+class MockCache:
+    """Mock for JSONFileCache to avoid touching files on disk"""
+
+    def __init__(self, working_dir=None, dumps_func=None):
+        self.working_dir = working_dir
+        self.dumps_func = dumps_func
+
+    def __contains__(self, cache_key):
+        return True
+
+    def __getitem__(self, cache_key):
+        return {
+            "startUrl": "https://test.awsapps.com/start",
+            "region": "us-east-1",
+            "accessToken": "access-token",
+            "expiresAt": TIME_IN_ONE_HOUR.strftime('%Y-%m-%dT%H:%M:%SZ'),
+            "expiresIn": 3600,
+            "clientId": "client-12345",
+            "clientSecret": "client-secret",
+            "registrationExpiresAt": TIME_IN_SIX_MONTHS.strftime(
+                '%Y-%m-%dT%H:%M:%SZ'
+            ),
+            "refreshToken": "refresh-here",
+        }
+
+    def __delitem__(self, cache_key):
+        pass
+
+
+class SSOSessionTest(BaseEnvVar):
+    def setUp(self):
+        super().setUp()
+        self.tempdir = tempfile.mkdtemp()
+        self.config_file = os.path.join(self.tempdir, 'config')
+        self.environ['AWS_CONFIG_FILE'] = self.config_file
+        self.access_key_id = 'ASIA123456ABCDEFG'
+        self.secret_access_key = 'secret-key'
+        self.session_token = 'session-token'
+
+    def tearDown(self):
+        shutil.rmtree(self.tempdir)
+        super().tearDown()
+
+    def write_config(self, config):
+        with open(self.config_file, 'w') as f:
+            f.write(config)
+
+    def test_token_chosen_from_provider(self):
+        profile = (
+            '[profile sso-test]\n'
+            'region = us-east-1\n'
+            'sso_session = sso-test-session\n'
+            'sso_account_id = 12345678901234\n'
+            'sso_role_name = ViewOnlyAccess\n'
+            '\n'
+            '[sso-session sso-test-session]\n'
+            'sso_region = us-east-1\n'
+            'sso_start_url = https://test.awsapps.com/start\n'
+            'sso_registration_scopes = sso:account:access\n'
+        )
+        self.write_config(profile)
+
+        session = Session(profile='sso-test')
+        with SessionHTTPStubber(session) as stubber:
+            self.add_credential_response(stubber)
+            stubber.add_response()
+            with mock.patch.object(
+                SSOTokenProvider, 'DEFAULT_CACHE_CLS', MockCache
+            ):
+                c = session.create_client('s3')
+                c.list_buckets()
+
+        self.assert_valid_sso_call(
+            stubber.requests[0],
+            (
+                'https://portal.sso.us-east-1.amazonaws.com/federation/credentials'
+                '?role_name=ViewOnlyAccess&account_id=12345678901234'
+            ),
+            b'access-token',
+        )
+        self.assert_credentials_used(
+            stubber.requests[1],
+            self.access_key_id.encode('utf-8'),
+            self.session_token.encode('utf-8'),
+        )
+
+    def test_mismatched_session_values(self):
+        profile = (
+            '[profile sso-test]\n'
+            'region = us-east-1\n'
+            'sso_session = sso-test-session\n'
+            'sso_start_url = https://test2.awsapps.com/start\n'
+            'sso_account_id = 12345678901234\n'
+            'sso_role_name = ViewOnlyAccess\n'
+            '\n'
+            '[sso-session sso-test-session]\n'
+            'sso_region = us-east-1\n'
+            'sso_start_url = https://test.awsapps.com/start\n'
+            'sso_registration_scopes = sso:account:access\n'
+        )
+        self.write_config(profile)
+
+        session = Session(profile='sso-test')
+        with pytest.raises(InvalidConfigError):
+            c = session.create_client('s3')
+            c.list_buckets()
+
+    def test_missing_sso_session(self):
+        profile = (
+            '[profile sso-test]\n'
+            'region = us-east-1\n'
+            'sso_session = sso-test-session\n'
+            'sso_start_url = https://test2.awsapps.com/start\n'
+            'sso_account_id = 12345678901234\n'
+            'sso_role_name = ViewOnlyAccess\n'
+            '\n'
+        )
+        self.write_config(profile)
+
+        session = Session(profile='sso-test')
+        with pytest.raises(InvalidConfigError):
+            c = session.create_client('s3')
+            c.list_buckets()
+
+    def assert_valid_sso_call(self, request, url, access_token):
+        assert request.url == url
+        assert 'x-amz-sso_bearer_token' in request.headers
+        assert request.headers['x-amz-sso_bearer_token'] == access_token
+
+    def assert_credentials_used(self, request, access_key, session_token):
+        assert access_key in request.headers.get('Authorization')
+        assert request.headers.get('X-Amz-Security-Token') == session_token
+
+    def add_credential_response(self, stubber):
+        response = {
+            'roleCredentials': {
+                'accessKeyId': self.access_key_id,
+                'secretAccessKey': self.secret_access_key,
+                'sessionToken': self.session_token,
+                'expiration': TIME_IN_ONE_HOUR.timestamp() * 1000,
+            }
+        }
+        stubber.add_response(body=json.dumps(response).encode('utf-8'))
